@@ -34,10 +34,14 @@ __all__ = [
 ]
 
 import abc
-from collections import deque
+from collections import (
+    defaultdict,
+    deque,
+)
+import logging
 from logging import getLogger
 from random import choice
-from select import select
+import selectors
 from socket import (
     AF_INET,
     AF_INET6,
@@ -141,6 +145,10 @@ class Bolt(abc.ABC):
         self.unresolved_address = unresolved_address
         self.socket = sock
         self.server_info = ServerInfo(Address(sock.getpeername()), self.PROTOCOL_VERSION)
+        # so far `connection.recv_timeout_seconds` is the only available
+        # configuration hint that exists. Therefore, all hints can be stored at
+        # connection level. This might change in the future.
+        self.configuration_hints = {}
         self.outbox = Outbox()
         self.inbox = Inbox(self.socket, on_error=self._set_defunct_read)
         self.packer = Packer(self.outbox)
@@ -331,7 +339,7 @@ class Bolt(abc.ABC):
             from neo4j.io._bolt4 import Bolt4x3
             connection = Bolt4x3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
         else:
-            log.debug("[#%04X]  S: <CLOSE>", s.getpeername()[1])
+            log.debug("[#%04X]  S: <CLOSE>", s.getsockname()[1])
             _close_socket(s)
 
             supported_versions = Bolt.protocol_handlers().keys()
@@ -498,7 +506,7 @@ class Bolt(abc.ABC):
 
     @abc.abstractmethod
     def fetch_message(self):
-        """ Receive at least one message from the server, if available.
+        """ Receive at most one message from the server, if available.
 
         :return: 2-tuple of number of detail messages and number of summary
                  messages fetched
@@ -536,7 +544,7 @@ class Bolt(abc.ABC):
         direct_driver = isinstance(self.pool, BoltPool)
 
         if error:
-            log.error(str(error))
+            log.debug("[#%04X] %s", self.socket.getsockname()[1], error)
         log.error(message)
         # We were attempting to receive data but the connection
         # has unexpectedly terminated. So, we need to close the
@@ -606,7 +614,7 @@ class IOPool:
         self.opener = opener
         self.pool_config = pool_config
         self.workspace_config = workspace_config
-        self.connections = {}
+        self.connections = defaultdict(deque)
         self.lock = RLock()
         self.cond = Condition(self.lock)
 
@@ -628,35 +636,52 @@ class IOPool:
             timeout = self.workspace_config.connection_acquisition_timeout
 
         with self.lock:
-            try:
-                connections = self.connections[address]
-            except KeyError:
-                connections = self.connections[address] = deque()
-
             def time_remaining():
                 t = timeout - (perf_counter() - t0)
                 return t if t > 0 else 0
 
             while True:
                 # try to find a free connection in pool
-                for connection in list(connections):
+                for connection in list(self.connections.get(address, [])):
                     if (connection.closed() or connection.defunct()
-                            or connection.stale()):
+                            or (connection.stale() and not connection.in_use)):
                         # `close` is a noop on already closed connections.
                         # This is to make sure that the connection is gracefully
                         # closed, e.g. if it's just marked as `stale` but still
                         # alive.
+                        if log.isEnabledFor(logging.DEBUG):
+                            log.debug(
+                                "[#%04X]  C: <POOL> removing old connection "
+                                "(closed=%s, defunct=%s, stale=%s, in_use=%s)",
+                                connection.local_port,
+                                connection.closed(), connection.defunct(),
+                                connection.stale(), connection.in_use
+                            )
                         connection.close()
-                        connections.remove(connection)
+                        try:
+                            self.connections.get(address, []).remove(connection)
+                        except ValueError:
+                            # If closure fails (e.g. because the server went
+                            # down), all connections to the same address will
+                            # be removed. Therefore, we silently ignore if the
+                            # connection isn't in the pool anymore.
+                            pass
                         continue
                     if not connection.in_use:
                         connection.in_use = True
                         return connection
                 # all connections in pool are in-use
-                infinite_pool_size = (self.pool_config.max_connection_pool_size < 0 or self.pool_config.max_connection_pool_size == float("inf"))
-                can_create_new_connection = infinite_pool_size or len(connections) < self.pool_config.max_connection_pool_size
+                connections = self.connections[address]
+                max_pool_size = self.pool_config.max_connection_pool_size
+                infinite_pool_size = (max_pool_size < 0
+                                      or max_pool_size == float("inf"))
+                can_create_new_connection = (
+                        infinite_pool_size
+                        or len(connections) < max_pool_size
+                )
                 if can_create_new_connection:
-                    timeout = min(self.pool_config.connection_timeout, time_remaining())
+                    timeout = min(self.pool_config.connection_timeout,
+                                  time_remaining())
                     try:
                         connection = self.opener(address, timeout)
                     except ServiceUnavailable:
@@ -699,7 +724,9 @@ class IOPool:
         """
         with self.lock:
             for connection in connections:
-                if not connection.is_reset:
+                if not (connection.is_reset
+                        or connection.defunct()
+                        or connection.closed()):
                     try:
                         connection.reset()
                     except (Neo4jError, DriverError, BoltError) as e:
@@ -1244,8 +1271,9 @@ def _handshake(s, resolved_address):
 
     # Handle the handshake response
     ready_to_read = False
-    while not ready_to_read:
-        ready_to_read, _, _ = select((s,), (), (), 1)
+    with selectors.DefaultSelector() as selector:
+        selector.register(s, selectors.EVENT_READ)
+        selector.select(1)
     try:
         data = s.recv(4)
     except OSError:
@@ -1298,6 +1326,14 @@ def connect(address, *, timeout, custom_resolver, ssl_context, keep_alive):
             s = _secure(s, resolved_address.host_name, ssl_context)
             return _handshake(s, resolved_address)
         except (BoltError, DriverError, OSError) as error:
+            try:
+                local_port = s.getsockname()[1]
+            except (OSError, AttributeError):
+                local_port = 0
+            err_str = error.__class__.__name__
+            if str(error):
+                err_str += ": " + str(error)
+            log.debug("[#%04X]  C: <CONNECTION FAILED> %s", local_port, err_str)
             if s:
                 _close_socket(s)
             errors.append(error)
